/*
  Theseus puzzle solver
  Copyright 2001 Eric Smith <eric@brouhaha.com>
 
  theseus is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.  Note that I am not
  granting permission to redistribute or modify theseus under the
  terms of any later version of the General Public License.
 
  This program is distributed in the hope that it will be useful
  (or at least amusing), but WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with these programs (in the file "COPYING"); if not, write to
  the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
  02139, USA.
*/

#include <stdio.h>

typedef int boolean;

typedef struct
{
  int row;
  int col;
} coord_t;

#define MAX_TURN 1000

#define MAZE_HEIGHT 9
#define MAZE_WIDTH 15

boolean horizontal_wall [MAZE_HEIGHT+1] [MAZE_WIDTH] =
{
  { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 },
  { 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1 },
  { 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 0, 1 },
  { 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0 },
  { 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0 },
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 },
  { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 },
};

boolean vertical_wall [MAZE_HEIGHT] [MAZE_WIDTH+1] =
{
  { 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0 },
  { 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1 },
  { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 },
  { 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0 },
  { 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0 },
  { 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0 },
  { 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0 },
  { 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0 },
  { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 }
};

boolean known_location [MAZE_HEIGHT] [MAZE_WIDTH] [MAZE_HEIGHT] [MAZE_WIDTH];

coord_t theseus;
coord_t minotaur;


void pchar2 (char c)
{
  printf ("%c%c", c, c);
}

void print_horiz_wall (int row)
{
  int col;
  for (col = 0; col <= MAZE_WIDTH; col++)
    {
      printf ("+");
      if (col < MAZE_WIDTH)
	pchar2 (" -" [horizontal_wall [row][col]]);
    }
  printf ("\n");
}

void print_row (int row)
{
  int col;
  for (col = 0; col <= MAZE_WIDTH; col++)
    {
      printf ("%c", " |" [vertical_wall [row][col]]);
      if (col < MAZE_WIDTH)
	{
	  if ((row == minotaur.row) && (col == minotaur.col))
	    pchar2 ('M');
	  else if ((row == theseus.row) && (col == theseus.col))
	    pchar2 ('T');
	  else
	    pchar2 (' ');
	}
    }
  printf ("\n");
}

void print_maze (void)
{
  int row;
  for (row = 0; row <= MAZE_HEIGHT; row++)
    {
      print_horiz_wall (row);
      if (row < MAZE_HEIGHT)
	print_row (row);
    }
}


void reset_locations (void)
{
  theseus.row = 4;
  theseus.col = 0;

  minotaur.row = 4;
  minotaur.col = 13;
}


void mark_location (void)
{
  known_location [theseus.row] [theseus.col] [minotaur.row] [minotaur.col] = 1;
}


boolean been_there_done_that (void)
{
  return (known_location [theseus.row] [theseus.col] [minotaur.row] [minotaur.col]);
}


typedef enum
{
  left, right, up, down, wait
} move_t;

char move_char [5] = { 'L', 'R', 'U', 'D', 'W' };



move_t theseus_moves [MAX_TURN];



void print_moves (int last_turn)
{
  int i;
  static int solutions = 0;
  printf ("solution %d, %d turns: ", ++solutions, last_turn + 1);
  for (i = 0; i <= last_turn; i++)
    printf ("%c", move_char [theseus_moves [i]]);
  printf ("\n");
}



boolean dead (void)
{
  return ((theseus.row == minotaur.row) && (theseus.col == minotaur.col));
}


boolean done (void)
{
  return ((theseus.row == 1) && (theseus.col == 14));
}


void minotaur_move (void)
{
  if ((minotaur.col > theseus.col) && ! vertical_wall [minotaur.row] [minotaur.col])
    {
      minotaur.col--;
      return;
    }
  if ((minotaur.col < theseus.col) && ! vertical_wall [minotaur.row] [minotaur.col+1])
    {
      minotaur.col++;
      return;
    }
  if ((minotaur.row > theseus.row) && ! horizontal_wall [minotaur.row] [minotaur.col])
    {
      minotaur.row--;
      return;
    }
  if ((minotaur.row < theseus.row) && ! horizontal_wall [minotaur.row+1] [minotaur.col])
    {
      minotaur.row++;
      return;
    }
}


void try_theseus_moves (int turn);


void theseus_move (int turn, move_t move)
{
  if (dead ())
    return;
  if (done ())
    {
      print_moves (turn);
      return;
    }
  theseus_moves [turn] = move;
  /* the minotaur gets two moves */
  minotaur_move ();
  if (dead ())
    return;
  minotaur_move ();
  if (dead ())
    return;
  if (been_there_done_that ())
    return;
  mark_location ();
  try_theseus_moves (turn+1);
}


void try_theseus_move_left (int turn)
{
  if (vertical_wall [theseus.row] [theseus.col])
    return;
  theseus.col--;
  theseus_move (turn, left);
}

void try_theseus_move_right (int turn)
{
  if (vertical_wall [theseus.row] [theseus.col+1])
    return;
  theseus.col++;
  theseus_move (turn, right);
}

void try_theseus_move_up (int turn)
{
  if (horizontal_wall [theseus.row] [theseus.col])
    return;
  theseus.row--;
  theseus_move (turn, up);
}

void try_theseus_move_down (int turn)
{
  if (horizontal_wall [theseus.row+1] [theseus.col])
    return;
  theseus.row++;
  theseus_move (turn, down);
}

void try_theseus_move_wait (int turn)
{
  theseus_move (turn, wait);
}


void try_theseus_moves (int turn)
{
  move_t move;
  coord_t old_theseus, old_minotaur;

#if 0
  printf ("turn %d:\n", turn);
  print_maze ();
#endif

  old_theseus = theseus;  old_minotaur = minotaur;

  try_theseus_move_left (turn);
  theseus = old_theseus;  minotaur = old_minotaur;

  try_theseus_move_right (turn);
  theseus = old_theseus;  minotaur = old_minotaur;

  try_theseus_move_up (turn);
  theseus = old_theseus;  minotaur = old_minotaur;

  try_theseus_move_down (turn);
  theseus = old_theseus;  minotaur = old_minotaur;

  try_theseus_move_wait (turn);
  theseus = old_theseus;  minotaur = old_minotaur;
}


int main (int argc, char *argv)
{
  reset_locations ();

  mark_location ();

  try_theseus_moves (0);
}
